/*
	preamp.c
	
	Copyright (C) 2002 Tim Goetze <tim@quitte.de>
	
	http://quitte.de/dsp/

	emulation of a 12AX7-based valve preamp. hard clipping (not
	antialiased) if the 'gain' parameter exceeds 1, soft harmonic
	distortion below.

	a 1996-sample ramp response from a spice model is interpolated
	linearly to find amplitude mappings.
*/
/*
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA or point your web browser to http://www.gnu.org.
*/

#include <malloc.h>
#include <ladspa.h>

#include "r12ax7.h"

typedef struct 
{
	LADSPA_Data * in;
	LADSPA_Data * out;

	LADSPA_Data * gain;

	LADSPA_Data adding_gain;
} 
preamp;

/* exp2 has been posted to music-dsp by andrew simper <andy@vellocet.com> 
 * a good pow (2, x) approximation, costing about 200 instead of 9000
 * cycles on a k6.
 * 
 * thanks, andy!
 */

inline float
exp2 (float x)
{
	long * px = (long *) (&x); // store address of float as long pointer
	const float tx = (x - .5f) + (3 << 22); // temporary value for truncation
	const long  lx = *((long *) &tx) - 0x4b400000; // integer power of 2
	const float dx = x - (float) (lx); // float remainder of power of 2
	x = 1.f + dx * (0.6960656421638072f + // cubic apporoximation of 2^x
						dx * (0.224494337302845f +  // for x in the range [0, 1]
						dx * (0.07944023841053369f)));
	*px += (lx << 23); // add integer power of 2 to exponent

	return x;
}

#define PROCESS(assignment) \
	register LADSPA_Data g = *((preamp *) h)->gain; \
	register LADSPA_Data ig; \
	register LADSPA_Data x; \
	\
	register LADSPA_Data * in = ((preamp *) h)->in; \
	register LADSPA_Data * out = ((preamp *) h)->out; \
	\
	register LADSPA_Data gain = *((preamp *) h)->gain; \
	register LADSPA_Data adding_gain = ((preamp *) h)->adding_gain; \
	\
	register int i; \
	register int r; \
	\
	if (g > 2.f) \
	{ \
		g = .5f * (g - 2.f); \
		ig = 1.f - (.125f * g); \
		g = exp2 (g + 4.f) - 15.f; \
	} \
	else \
	{ \
		g *= .5f; \
		if (g < 0.0001f) \
			g = 0.0001f; \
		ig = g > .5f ? 1.f : .5f / g; \
	} \
	\
	for (i = 0; i < frames; ++i) \
	{ \
		x = g * in[i]; \
		\
		x = ((x + 1.f) / 2.f) * 1996.f; \
		\
		if (x < 0) x = 0.f; \
		if (x >= 1995.f) x = 1994.99999999f; \
		\
		r = (int) x; \
		x -= r; \
		\
		x = ig * (r12ax7[r] * (1.f - x) + r12ax7[r + 1] * x); \
		\
		assignment; \
	}

/* /////////////////////////////////////////////////////////////////////// */

static LADSPA_Handle
instantiate (const struct _LADSPA_Descriptor * desc, unsigned long sr)
{
	preamp * h = (preamp *) malloc (sizeof (preamp));
	h->adding_gain = 1.f;
	return h;
}

static void
connect_port (LADSPA_Handle h, unsigned long n, LADSPA_Data * to)
{
	if (n == 0)
		((preamp *) h)->in = to;
	else if (n == 1)
		((preamp *) h)->gain = to;
	else if (n == 2)
		((preamp *) h)->out = to;
}

static void
activate (LADSPA_Handle h)
{
	;
}

static void
run (LADSPA_Handle h, unsigned long frames)
{
	PROCESS (out[i] = x);
}

static void
run_adding (LADSPA_Handle h, unsigned long frames)
{
	PROCESS (out[i] += adding_gain * x);
}

static void
set_run_adding_gain (LADSPA_Handle h, LADSPA_Data adding_gain)
{
	((preamp *) h)->adding_gain = adding_gain;
}

static void
cleanup (LADSPA_Handle h)
{
	free (h);
}

/* /////////////////////////////////////////////////////////////////////// */

static const LADSPA_PortDescriptor 
port_descriptors[] = 
{
	LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
	LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL,
	LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
};

static const char *
port_names[] =
{
	"in",
	"gain",
	"out",
};

static const LADSPA_PortRangeHint
port_range_hints[] =
{
	/* gaan we mieren neuken? ja wel! */
	{
		LADSPA_HINT_BOUNDED_BELOW | 
		LADSPA_HINT_BOUNDED_ABOVE | 
		LADSPA_HINT_DEFAULT_MIDDLE,
		-1.0, 1.0
	},
	{
		LADSPA_HINT_BOUNDED_BELOW | 
		LADSPA_HINT_BOUNDED_ABOVE | 
		LADSPA_HINT_DEFAULT_1,
		0.0, 10.0
	},
	{
		LADSPA_HINT_BOUNDED_BELOW | 
		LADSPA_HINT_BOUNDED_ABOVE | 
		LADSPA_HINT_DEFAULT_MIDDLE,
		-1.0, 1.0
	},
};

static const LADSPA_Descriptor
descriptor =
{
	UniqueID:	1763,
	Label: "preamp",
	Properties: LADSPA_PROPERTY_HARD_RT_CAPABLE,
	Name: "12AX7 preamp (no antialiasing)",
	Maker: "Tim Goetze <tim@quitte.de>",
	Copyright: "GPL, (C) 2002 <tim@quitte.de>",
	
	PortCount: 3,
	PortDescriptors: port_descriptors,
	PortNames: port_names,
	PortRangeHints: port_range_hints,
	
	/* methods */
	instantiate: instantiate,
	connect_port: connect_port,
	activate: activate,
	run: run,
	run_adding: run_adding,
	set_run_adding_gain: set_run_adding_gain,
	cleanup: cleanup,
};

/* /////////////////////////////////////////////////////////////////////// */

const LADSPA_Descriptor *
ladspa_descriptor (unsigned long index) 
{
	return index ? 0 : &descriptor;
}

